from typing import Dict, List
from orca.core.errors import ConfigurationError


class OrcaTask(object):

    def __init__(self, task_dict: Dict, task_locals: Dict):
        self.task_data = task_dict
        self._status = 'unknown'
        self._task_locals = task_locals

    @property
    def name(self) -> str:
        try:
            return self.task_data.get('task')
        except KeyError as e:
            raise ConfigurationError("Missing Task Identifier: ", e)

    @property
    def locals(self) -> Dict:
        """The resolved inputs and generated ouputs"""
        return self._task_locals

    @property
    def inputs(self) -> Dict:
        """The inputs as found in the yaml file"""
        return self.task_data.get('inputs', {})

    @property
    def outputs(self) -> List:
        """The outputs as found in the yaml file"""
        return self.task_data.get('outputs', [])

    @property
    def config(self) -> Dict:
        return self.task_data.get('config', {})

    @property
    def csip(self) -> str:
        return self.task_data.get('csip')

    @property
    def http(self) -> str:
        return self.task_data.get('http')

    @property
    def bash(self) -> str:
        return self.task_data.get('bash')

    @property
    def python(self) -> str:
        return self.task_data.get('python')

    @property
    def status(self) -> str:
        return self._status

    @status.setter
    def status(self, st: str):
        self._status = st



import os
import subprocess as subp
import requests
import json
import re
from csip import Client
from typing import List, Dict
from orca.core.tasks import OrcaTask
from orca.core.config import task, log, OrcaConfig
from orca.core.ledger import Ledger
from concurrent.futures import ThreadPoolExecutor
from dotted.collection import DottedCollection
from abc import ABCMeta, abstractmethod
from orca.core.errors import ConfigurationError, ExecutionError
from orca.core.config import var  # noqa: F401


# some global utility functions


def values_tostr(d: Dict) -> Dict:
    return {key: str(value) for key, value in d.items()}


def handle_service_result(response: Dict, outputs, name: str) -> Dict:
    d = {}
    for k, v in response.items():
        log.debug(" handling req result props: {0} -> {1}".format(k, str(v)))
        if k in outputs:
            d[name + "." + k] = v
    return d


def handle_csip_result(response: Dict, outputs: List, name: str) -> Dict:
    d = {}
    for k, v in response.items():
        if k in outputs:
            try:
                d[name + "." + k] = v['value']
            except KeyError as e:
                raise ExecutionError(
                    "Output variable {0} not found, message: {1}," +
                    "\n the response from the task was: {2}".format(k, e, json.dumps(response, indent=2)))
    return d


def handle_python_result(outputs: List, name: str, task_locals: Dict) -> Dict:
    d = {}
    for out in outputs:
        try:
            d[name + '.' + out] = task_locals[out]
        except KeyError as e:
            raise ExecutionError("Task output not found: {0}".format(out), e)
    return d


#############################################

class OrcaHandler(metaclass=ABCMeta):
    """ Abstract orca handler for control flow, variable resolution"""

    def __init__(self):
        self.symtable = {}

    def _check_symtable(self, name: str, task: Dict):
        if name is None or not name.isidentifier():
            raise ConfigurationError('Invalid task name: "{0}"'.format(name))
        task_id = id(task)
        # check against the task dict id to support loops.
        if self.symtable.get(name, task_id) != task_id:
            raise ConfigurationError("Duplicate task name: {0}".format(name))
        self.symtable[name] = task_id

    def handle(self, config: OrcaConfig) -> None:
        self.config = config
        self._handle_sequence(config.job)
        self.close()

    def _handle_sequence(self, sequence: Dict) -> None:
        for step in sequence:
            node = next(iter(step))
            if node == "task":
                log.debug(" ---- task: '{}'".format(step['task']))
                self._handle_task(step)
            elif node == "if":
                log.debug(" ---- if: '{}'".format(step['if']))
                self._handle_if(step)
            elif node == "for":
                log.debug(" ---- for: '{}'".format(step['for']))
                self._handle_for(step)
            elif node.startswith("fork"):
                log.debug(" ---- fork: ")
                self._handle_fork(step['fork'])
            elif node == "switch":
                log.debug(" ---- switch: '{}'".format(step['switch']))
                self._handle_switch(step)
            else:
                raise ConfigurationError(
                    'Invalid step in job: "{0}"'.format(node))

    def close(self):
        pass

    # to be overwritten by subclasses
    @abstractmethod
    def handle_csip(self, task: OrcaTask) -> Dict:
        pass

    @abstractmethod
    def handle_http(self, task: OrcaTask) -> Dict:
        pass

    @abstractmethod
    def handle_bash(self, task: OrcaTask) -> Dict:
        pass

    @abstractmethod
    def handle_python(self, task: OrcaTask) -> Dict:
        pass

    def __select_handler(self, task_dict: Dict):
        if 'csip' in task_dict:
            return self.handle_csip
        elif 'http' in task_dict:
            return self.handle_http
        elif 'bash' in task_dict:
            return self.handle_bash
        elif 'python' in task_dict:
            return self.handle_python
        else:
            raise ConfigurationError(
                'Invalid task type: "{0}"'.format(task_dict))

    def resolve_task_inputs(self, task_dict: Dict) -> Dict:
        inputs = task_dict.get('inputs', {})
        if inputs:
            return {k: eval(str(v), globals()) for k, v in inputs.items()}
        return {}

    def _resolve_file_path(self, name: str, ext: str) -> str:
        """ resolve the full qualified path name"""

        # potential value resolution if this should be a file name
        try:
            name = eval(str(name), globals())
        except Exception as e:
            # ok, never mind
            log.debug(e)

        if os.path.isfile(name):
            return name
        else:
            # maybe change this, because of testing
            if hasattr(self, 'config'):
                yaml_dir = self.config.get_yaml_dir()
            else:
                yaml_dir = "."
            rel_path = os.path.join(yaml_dir, name)
            if os.path.isfile(rel_path):
                # path relative to yaml file
                return rel_path
            else:
                if name.endswith(ext):
                    # this should be a file but it's not.
                    raise ConfigurationError(
                        'File not found: "{0}"'.format(name))
                return None

    def _handle_task(self, task_dict: Dict) -> OrcaTask:
        name = task_dict.get('task', None)

        # check the symbol table for task name to be an unique and valid name
        self._check_symtable(name, task_dict)

        # task_locals are the resolved inputs, they will be used for
        # execution
        task_locals = self.resolve_task_inputs(task_dict)
        _task = OrcaTask(task_dict, task_locals)

        log.debug("task '{0}' locals pre: {1}".format(_task.name, task_locals))

        # select the handler and call handle
        handle = self.__select_handler(task_dict)
        log.info('Starting task {0}'.format(name))
        result = handle(_task)
        log.info('Task {0} completed'.format(name, ))

        log.debug("task '{0}' locals post: {1}".format(
            _task.name, task_locals))

        # put the task_locals into the global task dictonary
        # this includes input and outputs
        task[_task.name] = {}
        for k, v in task_locals.items():
            task[_task.name][k] = v

        return _task

    # control structures
    def _handle_if(self, condition_block: Dict) -> None:
        """Handle 'if'"""
        cond = condition_block['if']
        sequence = condition_block['do']
        if eval(cond, globals()):
            self._handle_sequence(sequence)

    def _handle_switch(self, condition_block: Dict) -> None:
        """Handle 'switch'"""
        cond = condition_block['switch']
        case = eval(cond, globals())
        seq = condition_block.get(case, condition_block.get('default', None))
        if seq is not None:
            self._handle_sequence(seq)

    def _handle_for(self, condition_block: Dict) -> None:
        """Handle 'for'"""
        var_expr = condition_block['for']
        i = var_expr.find(",")
        if i == -1:
            raise ConfigurationError(
                'Invalid "for" expression: "{0}"'.format(var_expr))

        var = var_expr[:i]
        if not var.isidentifier():
            raise ConfigurationError(
                'Not a valid identifier: "{0}"'.format(var))

        expr = var_expr[i + 1:]
        for i in eval(expr, globals()):
            # mapping loop variable 'i' to 'var'
            q = ''
            if isinstance(i, str):
                q = "'"
            exec("{0}={2}{1}{2}".format(var, i, q), globals())
            self._handle_sequence(condition_block['do'])

    def _handle_fork(self, sequences: Dict) -> None:
        """Handle 'fork'"""
        with ThreadPoolExecutor(max_workers=(len(sequences))) as executor:
            for sequence in sequences:
                executor.submit(self._handle_sequence, sequence)


#############################################

class ExecutionHandler(OrcaHandler):
    """Execution Handler, executes csip, bash, python, http"""

    def __init__(self, ledger: Ledger = None):
        super().__init__()
        self.ledger = ledger or Ledger()
        self.validator = ValidationHandler()

    def handle(self, config: OrcaConfig) -> None:
        log.info('Executing workflow...')
        self.ledger.set_config(config)
        super().handle(config)

    def close(self) -> None:
        self.ledger.close()

    def _handle_task(self, task_dict: Dict) -> None:
        _task = super()._handle_task(task_dict)
        self.ledger.add(_task)

    def handle_csip(self, _task: OrcaTask) -> Dict:
        try:
            client = Client()
            for key, value in _task.locals.items():
                if isinstance(value, DottedCollection):
                    client.add_data(key, value.to_python())
                else:
                    client.add_data(key, value)
            client = client.execute(_task.csip)
            return handle_csip_result(client.data, _task.outputs, _task.name)
        except requests.exceptions.HTTPError as e:
            raise ExecutionError(e)

    def handle_http(self, _task: OrcaTask) -> Dict:
        url = _task.http
        name = _task.name
        inputs = _task.locals
        headers = _task.config.get('header')
        content_type = headers.get('content-type', 'text/plain')

        if 'method' not in _task.config:
            raise ConfigurationError(
                "requests service operator must include method: service {0}".format(name))

        if _task.config.get('method') == 'GET':
            return handle_service_result(json.loads(requests.get(url, params=_task.config.get('params', None)).content),
                                         _task.outputs, name)
        elif _task.config['method'] == 'POST':
            if isinstance(inputs, DottedCollection):
                return handle_service_result(json.loads(requests.post(url, inputs.to_python()).content), _task.outputs,
                                             name)
            else:
                return handle_service_result(json.loads(requests.post(url, inputs).content), _task.outputs, name)

    def handle_bash(self, _task: OrcaTask) -> Dict:
        env = {}
        config = _task.config
        # get defaults
        delimiter = config.get('delimiter', '\n')
        wd = config.get('wd', None)
        if len(_task.locals) > 0:
            env = values_tostr(_task.locals)
        sp = subp.Popen(_task.bash, env=env, shell=True,
                        stdout=subp.PIPE, stderr=subp.PIPE, cwd=wd)
        out, err = sp.communicate()
        if sp.returncode != 0:
            log.error('return code: {0}'.format(sp.returncode))
        if err:
            for line in err.decode('utf-8').split(delimiter):
                log.error("ERROR: " + line)
        if out:
            o = ""
            for line in out.decode('utf-8').split(delimiter):
                if line:
                    o += line + '\n'
            return o
        return {}

    def __get_call_string(self, config: Dict, script: str):
        func_name = config.get('callable')
        var_name = config.get('returns', '')
        # match against a function definition string : def <funcname> ( args,...)
        pattern = r'((?P<keyword>def)\s?(?P<function>\w+)\s?\((?P<args>(?P<arg>\w+(,\s?)?)+)\))'
        # find all functions in the file
        all_funcs = re.findall(pattern, script)
        # take each function string and break it up into a dictionary so we can easily extract the arguments
        dicts = [re.match(pattern, func[0]).groupdict() for func in all_funcs]
        # filter the list of functions for the function the user has defined
        fn_dict = [d for d in dicts if d.get('function') == func_name][0]
        # make the string that will be eval'd
        assignment = var_name if var_name == '' else var_name + ' = '
        return '{0}{1}({2})'.format(assignment, func_name, fn_dict.get('args', ''))

    def handle_python(self, _task: OrcaTask):
        log.debug("  exec python file : " + _task.python)

        resolved_file = self._resolve_file_path(_task.python, ".py")
        config = _task.config
        try:
            if resolved_file is None:
                exec(_task.python, _task.locals)
            else:
                with open(resolved_file, 'r') as script:
                    _python = script.read()
                    exec(_python, _task.locals)
                    if 'callable' in config:
                        call_str = self.__get_call_string(config, _python)
                        exec(call_str, _task.locals)

            _task.status = "success"
        except IndexError:
            raise ExecutionError(
                'The function {0} was not defined in the file: {1}'.format(_task.config.get('callable'), _task.python)
            )
        except BaseException as e:
            _task.status = "failed"
            log.debug(str(e))
            raise
        # remove after execution
        keys_to_remove = [k for k in _task.locals if
                          k not in _task.outputs and k not in _task.inputs]

        for key in keys_to_remove:
            del _task.locals[key]


#############################################

class ValidationHandler(OrcaHandler):
    """ValidationHandler, no execution"""

    def __init__(self):
        super().__init__()

    def handle_csip(self, _task: OrcaTask):
        r = requests.head(_task.csip)
        if r.status_code >= 400:
            raise ConfigurationError(
                'Task {0} defines a CSIP endpoint that is not accessible: "{1}"'.format(_task.name, _task.csip)
            )

    def handle_http(self, _task: OrcaTask):
        r = requests.head(_task.http)
        if r.status_code >= 400:
            raise ConfigurationError(
                'Task {0} defines a Http url that is not accessible: "{1}"'.format(_task.name, _task.http)
            )

    def handle_bash(self, _task: OrcaTask):
        try:
            self._resolve_file_path(_task.bash, ".sh")
        except ConfigurationError as e:
            raise ConfigurationError(
                'Task {0} defines a bash script that cannot be found: {1}'.format(_task.name, _task.bash), e
            )

    def handle_python(self, _task: OrcaTask):
        try:
            self._resolve_file_path(_task.python, ".py")
        except ConfigurationError as e:
            raise ConfigurationError(
                'Task {0} defines a python script that cannot be found: {1}'.format(_task.name, _task.python), e
            )


#############################################

class NoneHandler(OrcaHandler):
    """Handler that does not do anything, useful for testing"""

    def __init__(self):
        super().__init__()

    def handle_csip(self, task: OrcaTask) -> Dict:
        pass

    def handle_http(self, task: OrcaTask) -> Dict:
        pass

    def handle_bash(self, task: OrcaTask) -> Dict:
        pass

    def handle_python(self, task: OrcaTask) -> Dict:
        pass


#############################################

class DotfileHandler(OrcaHandler):
    """Handles printing of a dot file"""

    def __init__(self):
        super().__init__()

    def handle(self, config: OrcaConfig) -> None:
        self.config = config
        self.dot = ['digraph {',
                    'START [shape=doublecircle,color=gray,fontsize=10]',
                    'END [shape=doublecircle,color=gray,fontsize=10]',
                    'node [style="filled",fontsize=10,fillcolor=aliceblue,color=gray,fixedsize=true]',
                    'edge [fontsize=9,fontcolor=dodgerblue3]'
                    ]
        self.last_task = 'START'
        self.last_vertex_label = ''
        # first pass: node declaration
        self.decl = True
        # unique id for each node
        self.idx = 0
        self._handle_sequence(config.job)
        self.dot.append('')
        # second pass: vertices
        self.decl = False
        self.idx = 0
        self._handle_sequence(config.job)
        self.close()

    def _handle_fork(self, sequences: Dict) -> None:
        """Handle parallel execution"""
        name = "fork_{0}".format(self.idx)
        term = "term_{0}".format(self.idx)
        self.idx += 1
        if self.decl:
            self.dot.append(
                '{0} [shape=house,fillcolor=cornsilk,fontcolor="dodgerblue3",label="FORK"]'.format(name))
            self.dot.append('{0} [shape=point]'.format(term))
            for sequence in sequences:
                self._handle_sequence(sequence)
        else:
            self.dot.append(
                '{0} -> {1} {2}'.format(self.last_task, name, self.last_vertex_label))
            self.last_vertex_label = ''
            for sequence in sequences:
                self.last_task = name
                self._handle_sequence(sequence)
                self.dot.append('{0} -> {1}'.format(self.last_task, term))
            self.last_task = term

    def _handle_for(self, conditional_block: Dict, ) -> None:
        """Handle Looping"""
        var_expr = conditional_block['for']
        sequence = conditional_block['do']
        name = "for_{0}".format(self.idx)
        term = "term_{0}".format(self.idx)
        self.idx += 1
        if self.decl:
            self.dot.append(
                '{0} [shape=trapezium,fillcolor=cornsilk,fontcolor="dodgerblue3",label="FOR\\n{1}"]'.format(name,
                                                                                                            var_expr))
            self.dot.append('{0} [shape=point]'.format(term))
            self._handle_sequence(sequence)
        else:
            self.dot.append(
                '{0} -> {1} {2}'.format(self.last_task, name, self.last_vertex_label))
            self.last_vertex_label = ''
            self.last_task = name
            self._handle_sequence(sequence)
            self.dot.append('{0} -> {1}'.format(self.last_task, term))
            self.dot.append('{0} -> {1} [style="dotted"]'.format(term, name))
            self.last_task = term

    def _handle_switch(self, conditional_block: Dict, ) -> None:
        """Handle conditional switch."""
        cond = conditional_block['switch']
        cases = conditional_block.copy()
        cases.pop('switch')
        name = "switch_{0}".format(self.idx)
        term = "term_{0}".format(self.idx)
        self.idx += 1
        if self.decl:
            self.dot.append(
                '{0} [shape=diamond,fillcolor=cornsilk,fontcolor="dodgerblue3",label="SWITCH\\n{1}"]'.format(name,
                                                                                                             cond))
            self.dot.append('{0} [shape=point]'.format(term))
            for case, seq in cases.items():
                self._handle_sequence(seq)
        else:
            self.dot.append(
                '{0} -> {1} {2}'.format(self.last_task, name, self.last_vertex_label))
            self.last_vertex_label = ''
            for case, seq in cases.items():
                self.last_task = name
                self.last_vertex_label = '[label="{0}"]'.format(case)
                self._handle_sequence(seq)
                self.dot.append('{0} -> {1}'.format(self.last_task, term))
            self.last_task = term

    def _handle_if(self, conditional_block: Dict) -> None:
        """Handle if."""
        cond = conditional_block['if']
        sequence = conditional_block['do']
        name = "if_{0}".format(self.idx)
        term = "term_{0}".format(self.idx)
        self.idx += 1
        if self.decl:
            self.dot.append(
                '{0} [shape=diamond,fillcolor=cornsilk,fontcolor="dodgerblue3",label="IF\\n{1}"]'.format(name, cond))
            self.dot.append('{0} [shape=point]'.format(term))
            self._handle_sequence(sequence)
        else:
            self.dot.append(
                '{0} -> {1} {2}'.format(self.last_task, name, self.last_vertex_label))
            self.last_task = name
            self.last_vertex_label = '[label="true"]'
            self._handle_sequence(sequence)
            self.dot.append('{0} -> {1}'.format(self.last_task, term))
            self.dot.append('{0} -> {1} [label="false"]'.format(name, term))
            self.last_task = term

    def close(self):
        self.dot.append("{0} -> END".format(self.last_task))
        self.dot.append("}")
        path, ext = os.path.splitext(self.config.get_yaml_file())
        with open(path + ".dot", "w") as text_file:
            print("\n".join(self.dot), file=text_file)
        log.info("generated dot file '" + path + ".dot'")

    def _ht(self, name: str, shape: str, label: str = ''):
        if self.decl:
            self.dot.append(
                '{0} [shape={1}, label="\'{0}\'\\n{2}"]'.format(name, shape, label))
        else:
            self.dot.append(
                '{0} -> {1} {2}'.format(self.last_task, name, self.last_vertex_label))
            self.last_vertex_label = ''
            self.last_task = name

    def handle_csip(self, task: OrcaTask):
        self._ht(task.name, 'cds', task.csip)

    def handle_http(self, task: OrcaTask):
        self._ht(task.name, 'cds', task.http)

    def handle_bash(self, task: OrcaTask):
        self._ht(task.name, 'note')

    def handle_python(self, task: OrcaTask):
        self._ht(task.name, 'note', task.python)



import unittest
from tests.mocks import tasks
from orca.core.tasks import OrcaTask
from orca.core.handler import ExecutionHandler, ValidationHandler
from orca.core.errors import ConfigurationError


class OrcaHandlerTest(unittest.TestCase):

    def test_inline_inputs_python(self):
        mock = tasks.inline_python_inputs_mock
        # _task = OrcaTask(mock, mock)
        h = ExecutionHandler()
        _task = OrcaTask(mock, h.resolve_task_inputs(mock))
        h.handle_python(_task)
        assert 'greeting' in _task.locals
        assert _task.locals['greeting'] == 'Hello Adam'

    def test_inline_python(self):
        mock = tasks.inline_python_mock
        h = ExecutionHandler()
        _task = OrcaTask(mock, h.resolve_task_inputs(mock))
        h.handle_python(_task)
        assert 'greeting' in _task.locals
        assert _task.locals['greeting'] == 'Hello World'

    def test_file_no_inputs_python(self):
        mock = tasks.file_python_mock
        h = ExecutionHandler()
        _task = OrcaTask(mock, h.resolve_task_inputs(mock))
        h.handle_python(_task)
        assert 'result' in _task.locals
        assert _task.locals['result'] == 10

    def test_file_inputs_python(self):
        mock = tasks.file_python_inputs_mock
        h = ExecutionHandler()
        _task = OrcaTask(mock, h.resolve_task_inputs(mock))
        h.handle_python(_task)
        assert 'result' in _task.locals
        assert _task.locals['result'] == 25

    def test_bad_file_python(self):
        mock = tasks.bad_file_path_python
        h = ValidationHandler()
        _task = OrcaTask(mock, h.resolve_task_inputs(mock))
        with self.assertRaises(ConfigurationError):
            h.handle_python(_task)




import orca as o   # must be renamed
from orca.core.config import OrcaConfig, log
from orca.core.handler import ExecutionHandler
from orca.core.handler import ValidationHandler
from orca.core.handler import DotfileHandler
from orca.core.ledger import JSONFileLedger
from orca.core.ledger import MongoLedger
from orca.core.ledger import KafkaLedger
from orca.core.errors import OrcaError

import click
import click_log

click_log.basic_config(log)


@click.group()
@click_log.simple_verbosity_option(log, default='INFO')
def orca():
    pass


@orca.command()
def version():
    """
    Print the orca version.
    """
    print(o.__version__)


def check_format(ctx, param, value):
    if value is None:
        return
    c = value.split('/')
    if len(c) != 3:
        log.error(
            "Invalid mongo connect string, expected '<host[:port]>/<db>/<col>'")
        ctx.exit()
    return c

def check_format_kafka(ctx, param, value):
    if value is None:
        return
    c = value.split('/')
    if len(c) != 2:
        log.error(
            "Invalid kafka connect string, expected '<host[:port]>/topic'")
        ctx.exit()
    return c


# run a workflow:
# python3 orca run --ledger-json /tmp/f.json for.yaml
# python3 orca run --ledger-mongo localhost/orcadb1/ledgercol for.yaml
#
# maybe import file into a db:
# mongoimport --db orca --collection ledger --file /tmp/f.json

@orca.command()
@click.option('--ledger-json', type=click.Path(), help='file ledger.')
@click.option('--ledger-mongo', type=str,
              help='mongodb ledger, TEXT format "<host[:port]>/<db>/<col>".', callback=check_format)
@click.option('--ledger-kafka', type=str,
              help='kafka ledger, TEXT format "<host[:port]>/topic".', callback=check_format_kafka)
@click.argument('file', type=click.File('r'))
@click.argument('args', nargs=-1)
def run(ledger_json, ledger_mongo, ledger_kafka, file, args):
    """
    Run an orca workflow.
    """
    try:
        ledger = None
        if ledger_json:
            ledger = JSONFileLedger(ledger_json)
        elif ledger_mongo:
            ledger = MongoLedger(ledger_mongo)
        elif ledger_kafka:
            ledger = KafkaLedger(ledger_kafka)

        config = OrcaConfig.create(file, args)
        executor = ExecutionHandler(ledger)
        executor.handle(config)
    except OrcaError as e:
        log.error(e)


@orca.command()
@click.argument('file', type=click.File('r'))
@click.argument('args', nargs=-1)
def validate(file, args):
    """
    Validate an orca workflow.
    """
    try:
        config = OrcaConfig.create(file, args)
        validator = ValidationHandler()
        validator.handle(config)
    except OrcaError as e:
        log.error(e)


# Create a visual workflow:
#    $ python orca todot abc.yaml
#  will generate 'abc.dot' in the same folder of abc.yaml
#  convert this with the 'dot' command:
#    $ dot -Tpdf switch.dot -o switch.pdf
#    $ dot -Tpng switch.dot -o switch.png
#    $ dot -Tsvg switch.dot -o switch.svg
@orca.command()
@click.argument('file', type=click.File('r'))
@click.argument('args', nargs=-1)
def todot(file, args):
    """
    Create a graphviz dot file from an orca workflow.
    """
    try:
        config = OrcaConfig.create(file, args)
        printer = DotfileHandler()
        printer.handle(config)
    except OrcaError as e:
        log.error(e)



import uuid
import os
import logging
import pymongo
import json

from orca.core.tasks import OrcaTask
from orca.core.config import OrcaConfig, log
from typing import Dict, List
from datetime import datetime
from pykafka import KafkaClient


class Ledger(object):
    """Keeps A record of workflow task executions transactions"""

    def __init__(self):
        # unique for each orca run
        self._id = uuid.uuid4()

    def set_config(self, config: OrcaConfig) -> None:
        self.config = config

    def add(self, task: OrcaTask) -> None:
        """Add an entry to the ledger"""
        # print("->> ledger: {0}".format(self._create_entry(task)))
        pass

    def _create_entry(self, task: OrcaTask) -> Dict:
        """Create a dictionary entry to record in a ledger"""
        d = {
            # the workflow file
            'orca_file': os.path.abspath(self.config.get_yaml_file()),
            # the 'version' entry in the workflow file
            'orca_id': self.config.get_version(),
            # the 'version' entry in the workflow file (e.g. gitattribute)
            'orca_name': self.config.get_name(),
            'task_name': task.name,  # the task name
            'task_uuid': str(self._id),  # the uuid of the current run
            'task_status': task.status,  # status of the run
            'task_time': str(datetime.now()),  # task execution time
            'task_data': task.locals,  # task data
        }
        if log.isEnabledFor(logging.DEBUG):
            log.debug('ledger: {0}'.format(d))
        return d

    def close(self) -> None:
        pass


############################################


class JSONFileLedger(Ledger):
    """Creates a JSON File with a ledger, appends if file exists."""

    def __init__(self, file: str):
        super().__init__()
        self.f = open(file, 'a+')
        log.debug('JSON Ledger: {0}'.format(self.f))

    def add(self, task: OrcaTask) -> None:
        self.f.write('{0}\n'.format(json.dumps(self._create_entry(task))))

    def close(self) -> None:
        self.f.close()
        log.debug('closed: {0}'.format(self.f))


############################################

class MongoLedger(Ledger):
    """Adds entries into a ledger hosted in Mongodb"""

    def __init__(self, con: List[str]):
        super().__init__()
        self.mongo = pymongo.MongoClient('mongodb://{0}/'.format(con[0]))
        self.col = self.mongo[con[1]][con[2]]
        log.debug('Mongo Ledger: {0} for {1}'.format(
            self.mongo, "/".join(con)))

    def add(self, task: OrcaTask) -> None:
        self.col.insert_one(self._create_entry(task))

    def close(self) -> None:
        self.mongo.close()
        log.debug('closed: {0}'.format(self.mongo))


#############################################


class KafkaLedger(Ledger):
    """ Kafka ledger"""

    def __init__(self, con: List[str]):
        super().__init__()
        self.kafka = KafkaClient(hosts=con[0])
        topic = self.kafka.topics[con[1]]
        self.producer = topic.get_producer()

    def add(self, task: OrcaTask) -> None:
        self.producer.produce(json.dumps(self._create_entry(task)))

    def close(self) -> None:
        self.kafka.close()
        log.debug('closed: {0}'.format(self.kafka))



from unittest import TestCase
from orca.core.errors import ConfigurationError
from tests.util import get_config


class OrcaSchemaTest(TestCase):
    pass

    def test_single_task(self):
        config = get_config('simple_example.yaml')
        assert config is not None

    def test_if_task_schema(self):
        config = get_config('if.yaml')
        assert config is not None

    def test_fork_task_schema(self):
        config = get_config('par.yaml')
        assert config is not None

    def test_switch_schema(self):
        config = get_config('switch.yaml')
        assert config is not None

    def test_for_schema(self):
        config = get_config('for.yaml')
        assert config is not None

    def test_duplicate_kinds(self):
        with self.assertRaises(ConfigurationError) as e:
            config = get_config('duplicate_kinds.yaml')
            assert config is not None
            print(e)

    def test_duplicate_outputs(self):
        with self.assertRaises(ConfigurationError) as e:

            config = get_config('duplicate_outputs.yaml')
            assert config is not None
            print(e)

    def test_nested_dups(self):
        with self.assertRaises(ConfigurationError) as e:
            config = get_config('nested_duplicates.yaml')
            assert config is not None
            print(e)

    def test_missing_api_version(self):
        with self.assertRaises(ConfigurationError):
            config = get_config('missing_api_version.yaml')
            assert config is not None

    def test_missing_api_version(self):
        with self.assertRaises(ConfigurationError):
            config = get_config('invalid_api_version.yaml')
            assert config is not None



import unittest
from orca.core.handler import ExecutionHandler
from orca.core.errors import ExecutionError
from tests.util import run_handler


class OrcaExecutionTest(unittest.TestCase):

    def test_comprehensive_example(self):
        run_handler('datetime.yaml', ExecutionHandler())

    def test_inline_bash_task(self):
        run_handler('bash-inline.yaml', ExecutionHandler())

    def test_csip_task(self):
        run_handler('csip.yaml', ExecutionHandler())

    def test_fork_task(self):
        run_handler('par.yaml', ExecutionHandler())

    def test_for_task(self):
        run_handler('for.yaml', ExecutionHandler())

    def test_switch_task(self):
        run_handler('switch.yaml', ExecutionHandler())

    def test_var1_task(self):
        run_handler('var1.yaml', ExecutionHandler())

    def test_var2_task(self):
        run_handler('var2.yaml', ExecutionHandler())

    def test_var3_task(self):
        run_handler('var3.yaml', ExecutionHandler())

    def test_var4_task(self):
        run_handler('var4.yaml', ExecutionHandler())

    def test_python_func(self):
        run_handler('python_funcs.yaml', ExecutionHandler())

    def test_bad_python_func(self):
        with self.assertRaises(ExecutionError):
            run_handler('bad_python_func.yaml', ExecutionHandler())


if __name__ == '__main__':
    unittest.main()



from click.testing import CliRunner
from orca.cli import commands
import unittest
from tests.util import get_config_path


class OrcaCommandTest(unittest.TestCase):

    def test_run(self):
        runner = CliRunner()
        runner.invoke(commands.run, [get_config_path('simple_example.yaml'), '--ledger-json=./test.json'])

    def test_version(self):
        runner = CliRunner()
        runner.invoke(commands.version)



import os

test_dir = os.path.dirname(os.path.dirname(__file__))
os.path.join(test_dir, 'fixtures', 'a.py')
inline_python_inputs_mock = {
    'task': 'inline_python',
    'python': 'greeting = \'Hello {0}\'.format(name)',
    'inputs': {
        'name': '"Adam"'
    },
    'outputs': [
        'greeting'
    ]
}

inline_python_mock = {
    'task': 'inline_python',
    'python': 'greeting = \'Hello World\'',
    'outputs': [
        'greeting'
    ]
}

file_python_inputs_mock = {
    'task': 'file_python',
    'python': os.path.join(test_dir, 'fixtures', 'inputs.py'),
    'inputs': {
        'input_1': 10,
        'input_2': 15
    },
    'outputs': [
        'result'
    ]
}

file_python_mock = {
    'task': 'file_python',
    'python': os.path.join(test_dir, 'fixtures', 'no_inputs.py'),
    'outputs': [
        'result'
    ]
}

bad_file_path_python = {
    'task': 'file_python',
    'python': os.path.join(test_dir, 'fixtures', 'does_not_exist.py'),
    'outputs': [
        'result'
    ]
}



from jsonschema import Draft7Validator, ValidationError
import logging
import os
import json
from ruamel import yaml
from typing import List, TextIO, Dict
from orca.core.errors import ConfigurationError
import itertools

log = logging.getLogger(__name__)

LATEST_SCHEMA_VERSION = '1.0'
AVAILABLE_SCHEMA_VERSIONS = ['1.0']


def _get_schema_location():
    """
    gets the dir of the schemas
    :return:  the schema directory
    """
    return os.path.dirname(os.path.abspath(__file__))


def _handle_errors(errors: List[ValidationError], fmt_err_func, filename):
    """
    Handles the tree of errors provided by jsonschema and formats them.
    :param errors:
    :param fmt_err_func:
    :param filename:
    :return:  exits if no errors exist
    """
    errors = list(sorted(errors, key=str))
    if not errors:
        return

    error_msg = '\n'.join(fmt_err_func(error) for error in errors)
    raise ConfigurationError(
        "The Orca configuration: {file} is invalid because:\n {error_msg}".format(
            file=" ' {}'".format(filename) if filename else "",
            error_msg=error_msg
        )
    )


def _dump_section(section: Dict):
    return yaml.dump(section)


def _parse_anyof_validator(error: ValidationError):
    """
    Parses a Jsonschema anyOf validator
    Most of our data is nested under anyOf validators, so we need reason about which validators are important and
    create meaningful error messages for them. Additionally we may need to recursively process anyOf validators until we
    find a meaningful error.
    :param error:  the toplevel Validation error
    :return:  path: the path to the error in the yaml tree,
    error_msg: the most relevant error_message for this branch of the tree.
    """

    def from_contexts(contexts):
        _errors = []
        for e in contexts:
            if e.validator == 'anyOf':
                _errors.extend(from_contexts(e.context))
            else:
                _errors.append(e)
        return _errors

    # recursively walk down the tree gathering nested exceptions deep in the workflow graph
    nested_errors = list(itertools.chain.from_iterable(
        [from_contexts(e.context) for e in error.context if e.validator == 'anyOf'])
    )
    # if we have nested errors, its likely that the real error is in there otherwise use top level errors
    errors = error.context if not nested_errors else nested_errors
    for error in errors:

        if error.validator == 'oneOf':
            return (error.absolute_path,
                    'The task "{0}" defines an incorrect kind, or more than one kind\n error message: {1}'.format(
                        error.instance.get('task'), error.message)
                    )

        if error.validator == 'required':
            return error.absolute_path, error.message

        if error.validator == 'additionalProperties':
            return error.absolute_path, error.message

        if error.validator == 'uniqueItems':
            return (error.absolute_path if error.path else None,
                    "contains non-unique items, please remove duplicates from {}".format(error.instance)
                    )

        if error.validator == 'type':
            return error.absolute_path, 'An invalid type was declared: {0}'.format(error.message)

        return error.absolute_path, error.message


def _handle_generic_error(error):
    """
    Handles a single ValidationError produced by jsonschema
    :param error: the error
    :return: a formatted error message
    """
    error_msg = error.message
    if error.validator == 'required':
        msg_format = 'missing a required property: {0}'
        return msg_format.format(error_msg)

    if error.validator == 'additionalProperties':
        msg_format = 'An invalid property has been defined: {0}'
        return msg_format.format(error_msg)

    if error.validator == 'type':
        msg_format = 'An invalid type was declared {0}'
        return msg_format.format(error_msg)

    if error.validator == 'anyOf':
        path, error_msg = _parse_anyof_validator(error)
        msg_format = 'Error validating job at {0}\n error message: {1}'
        return msg_format.format(path, error_msg)


def validate(file: TextIO):
    """
    Checks the orca file against a schema. which schema is determined by the 'apiVersion' defined in the
    orca configuration, if no configuration
    :param file:
    :return:
    """
    data = file.read()
    log.debug("Raw yaml: {0}".format(data))

    orca_data = yaml.load(data, yaml.Loader)
    try:
        version = orca_data['apiVersion']
    except KeyError:
        raise ConfigurationError(
            "'apiVersion is missing. An API version must be specified on an orca document," +
            " the latest current version is {0}'".format(LATEST_SCHEMA_VERSION)
        )

    schema_file = os.path.join(_get_schema_location(), "ORCA_SCHEMA_{0}.json".format(version))
    try:  # check if the schema file exists.
        with open(schema_file) as fp:
            schema_data = json.load(fp)
            validator = Draft7Validator(schema_data)
            errors = list(validator.iter_errors(orca_data))
            _handle_errors(errors, _handle_generic_error, file.name)
            return data
    except FileNotFoundError:
        raise ConfigurationError(
            "'The apiVersion {0}, is an invalid apiVersion version. It did not match one of the".format(version) +
            " supported apiVersions: {0}'".format(AVAILABLE_SCHEMA_VERSIONS)
        )



import json
import logging
import os
import re
from typing import List, Dict, TextIO

from dotted.collection import DottedDict
from ruamel import yaml
from orca.schema.validation import validate
from orca.core.errors import ConfigurationError
log = logging.getLogger(__name__)

# all payload data during processing. must be global!
task = DottedDict()
var = DottedDict()


class OrcaConfig(object):
    """ Orca configuration class"""

    @staticmethod
    def __process_config(file: TextIO) -> Dict:
        try:
            # first pass: start by validating the yaml file against the schema version.
            data = validate(file)
            # processing single quote string literals: " ' '
            repl = r"^(?P<key>\s*[^#:]*):\s+(?P<value>['].*['])\s*$"
            fixed_data = re.sub(repl, '\g<key>: "\g<value>"',
                                data, flags=re.MULTILINE)
            log.debug("Processed yaml: {0}".format(fixed_data))
            # second pass: appropriately quote strings in the yaml file.
            config = yaml.load(fixed_data, Loader=yaml.Loader)

            if log.isEnabledFor(logging.DEBUG):  # to avoid always dump json
                log.debug("Loaded yaml: {0}".format(
                    json.dumps(config, indent=2)))

            return config
        except yaml.YAMLError as e:
            log.error(e)
            raise ConfigurationError("error loading yaml file.", e)
        except ConfigurationError as e:
            # lets capture it log it and reraise it.
            log.error(e)
            raise e

    @staticmethod
    def create(file: TextIO, args: List[str] = None) -> 'OrcaConfig':
        d = OrcaConfig.__process_config(file)
        return OrcaConfig(d, file.name, args)

    def __init__(self, config: Dict, file: str = None, args: List[str] = None):
        # the yaml file (if used)

        self.file = file
        self.api_version = config.get('apiVersion')
        self.conf = config.get('conf', {})
        self.deps = config.get('dependencies', [])
        self.var = config.get('var', {})
        self.job = config.get('job')
        self.version = config.get('version', '0.0')
        self.name = config.get('name', file)

        self.__set_vars({} if self.var is None else self.var,
                        args if args is not None else [])

    def get_yaml_dir(self) -> str:
        return os.path.dirname(self.file) if self.file is not None else "."

    def get_yaml_file(self) -> str:
        return self.file

    def get_version(self) -> str:
        return self.version

    def get_name(self) -> str:
        return self.name

    def __resolve_dependencies(self):
        for dep in self.deps:
            try:
                exec("import " + dep, globals())
                log.debug("importing dependency: '{0}'".format(dep))
            except Exception as e:
                raise ConfigurationError(
                    "Cannot not resolve the '{0}' dependency".format(dep), e)

    def __set_vars(self, variables: Dict, args: List[str]) -> None:
        """put all variables as globals"""
        log.debug("setting job variables:")
        for key, val in variables.items():
            if not key.isidentifier():
                raise ConfigurationError(
                    'Invalid variable identifier: "{0}"'.format(key))
            try:
                exec("var.{0}={1}".format(key, val), globals())
                log.debug(
                    "  set var.{0} = {1} -> {2}".format(key, str(val), str(eval("var." + key))))
            except Exception as e:
                raise ConfigurationError(
                    "Cannot set variable: {0}".format(key), e)



import setuptools
import orca

setuptools.setup(
    name=orca.name,
    version=orca.__version__,
    author="Adam Shelton, Olaf David",
    author_email="adamshelton@rti.org",
    install_requires=[
        "csip ==  0.8.13",
        "jsonschema == 3.0.1",
        "click == 7.0",
        "click-log == 0.3.2",
        "dotted == 0.1.8",
        "ruamel.yaml == 0.15.88",
        "pymongo == 3.7.2",
        "requests == 2.21.0",
        "pykafka == 2.8.0" 
    ],
    description="A cli tool for orchestrating model workflows",
    long_description=open('README.md').read(),
    url="https://github.com/KoduIsGreat/orca.git",
    include_package_data=True,
    package_data={'orca.schema': ['*.json']},
    packages=setuptools.find_packages(exclude=['tests']),
    entry_points={
        'setuptools.installation': [
            'eggsecutable = orca.cli.commands:orca'
        ],
        'console_scripts': [
            'orca=orca.cli.commands:orca'
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers"
    ]
)



import unittest
from orca.core.handler import DotfileHandler

from tests.util import run_handler


class DotFileHandlerTests(unittest.TestCase):

    def test_print_switch_task(self):
        run_handler('switch.yaml', DotfileHandler())

    def test_print_for_task(self):
        run_handler('for.yaml', DotfileHandler())

    def test_print_par_task(self):
        run_handler('par.yaml', DotfileHandler())


if __name__ == '__main__':
    unittest.main()
